// --- [ RS5C372A.c ] --------------------------------------------------------
//
// tab = 3
//
// Funktionen fr die RealTimeClock Ricoh RS5C372A
//
// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "global.h"
#include "TWI_Master.h"
#include "RS5C372A.h"
#include "rs232.h"


// lokale Funktionen
uint8_t bcd_2_dez(uint8_t bcd);
uint8_t dez_2_bcd(uint8_t dez);

void clrRegisters(void);
void getCtrlRegisters(uint8_t *);
void setRegisters(uint8_t startReg, uint8_t * values, uint8_t cnt);
void error(uint8_t);

// lokale Variable fr den aktuellen Wert des ControlReg 1
// mit den Enable-Flags fr Alarm
uint8_t ctrlReg_1 = 0;

// lokales Array fr die TWI-Kommunikation (16 Byte fr die Register + TWI-Adresse + RegisterAdresse)
uint8_t twibuf[18];  			// fr die TWI-Kommunikation
uint8_t rtc_data[16]; 			// zum Vorbereiten von zu sendenden Datenpaketen

// ----------------------------------------------------------------------------
// lokal -> liest die Inhalte der CTRL-Register in lokale Variablen,
// Prft dann, ob im Register CTRL_REG_2 (0x0F) das Bit XSTP gesetzt ist:
// wenn ja, dann liegt ein Neustart vor,
// - es werden alle Register auf die Default-Werte gesetzt.
// wenn nein, dann wird das ControlReg 1 in eine lokale Variable eingelesen.
void rtc_init(void)
{
	uint8_t r[2];
	getCtrlRegisters(r);					// alle 16 Register, beginnend mit 0x0F einlesen

	put_c(13);
	puts_p(PSTR("CTRL_1: "));
	print_hexbyte(r[0]);
	puts_p(PSTR(" CTRL_2: "));
	print_hexbyte(r[1]);
	put_c(13);

	if (r[1] & (1<<XSTP))				// CTRL_REG_2, PowerDown
	{
		error(CONTINUE);					// unkritischer Fehler
	}

	else
	{
		ctrlReg_1 = r[0] & 0xC7;		// Die AlarmEnables beibehalten, CtrlReg_1 im Speicher halten
		rtc_data[0] = ctrlReg_1;		//
		rtc_data[1] = (1<<HOUR24 | 1<<CLEN);	// Standard-Wert fr CtrlRegister 2
		setRegisters(CTRL_REG_1, rtc_data, 2); // CTRL_REG_1 und CTLR_REG_2 schreiben

		rtc_data[0] = TIME_TRIM;					//
		setRegisters(TRIM_REG, rtc_data, 1);	// den Trimmwert schreiben

		puts_p(PSTR("RTC ok"));
		put_c(13);
	}
}

// ----------------------------------------------------------------------------
// lokal -> Schreibt die default-Werte / 0 in die Register
//
void clrRegisters(void)
{
	uint8_t i;
														// ControlReg 2
	rtc_data[0] = (1<<HOUR24 | 1<<CLEN);  	// 24-Stunden Modus, 32KHz Ausgang aus
	for (i = 1; i < 16; i++)
	{
		rtc_data[i] = 0;							// alle Register auf 0 setzen
	}
	rtc_data[TRIM_REG] = TIME_TRIM;			// aber noch den Trimwert einfgen

	setRegisters(CTRL_REG_2, rtc_data, 16);
}

// ----------------------------------------------------------------------------
// Schreibt die gesamte Zeitinformation auf die RTC
// geschreiben wird ab Register 0
// time[] dezimal in der Reihenfolge [Sekunde, Minute, Stunde, DOW, Tag, Monat, Jahr)
void rtc_setTime(uint8_t * time)
{
	uint8_t i;

	// die Zeit dezimal umwandeln in BCD-Zahlen
	for (i = 0; i < 7; i++)
		{
			// Der Sonderfall DOW kann bei der Konvertierung ignoriert werden, da immer < 7
			// daher richtet die Konvertierung keinen Schaden an.
			rtc_data[i] = dez_2_bcd(time[i]);
		}
	setRegisters(0, rtc_data, 7);
}

// ----------------------------------------------------------------------------
// Holt die Zeit aus der RTC ab und schreibt sie ins Array time[]
// time[] dezimal in der Reihenfolge [Sekunde, Minute, Stunde, DOW, Tag, Monat, Jahr)
// Gelesen wird ab Register 0x0F !!!
// In diesem Register wird auch das Bit XSTP geprft,
// wenn es gesetzt ist, dann hat es einen PowerDown gegeben -
// dann wird ein Init() ausgefhrt.
void rtc_getTime(uint8_t * time)
{
	uint8_t i, j;

	twibuf[0] = RTC_TWI_ADR;
	TWI_MA_Read(twibuf, 9);				// CtrlReg_2 + 7 Byte lesen
												// twibuf[0] = TWI_Adr
	if (twibuf[1] & (1<<XSTP))			// keine gltige Zeitinformation wegen PowerDown
	{
		// auf den Fehler reagieren ....
		error(QUIT);
	}

	else
	{
		j = 0;
		// die Zeit von BCD nach DEZ wandeln
		for (i = 2; i < 9; i++)
		{
			// Der Sonderfall DOW kann hier ignoriert werden, da immer < 7
			// daher richtet die Konvertierung keinen Schaden an.
			time[j++] = bcd_2_dez(twibuf[i]);
		}
	}
}

// ----------------------------------------------------------------------------
// Setzt den Alarm_A - Parameter: Minute, Stunde, DOW (0..6)
// und setzt das Flag AALE im Ctrl-Register 1
void rtc_setAlarm_A(uint8_t * time)
{
	ctrlReg_1 |= (1<<AALE);						// AlarmFlag_A in ctrlReg_1 setzen
	rtc_data[0] = ctrlReg_1;
	setRegisters(CTRL_REG_1, rtc_data, 1); // und schreiben

	rtc_data[0] = dez_2_bcd(time[0]);
	rtc_data[1] = dez_2_bcd(time[1]);
	rtc_data[2] = time[2];						// Format DOW muss dezimal bleiben !

	setRegisters(ALARM_REG_A, rtc_data, 3);
}

// ----------------------------------------------------------------------------
// Setzt den Alarm_B - Parameter: Minute, Stunde, DOW (0..6)
// und setzt das Flag BALE im Ctrl-Register 1
void rtc_setAlarm_B(uint8_t * time)
{
	ctrlReg_1 |= (1<<BALE);						// AlarmFlag_B in ctrlReg_2 setzen
	rtc_data[0] = ctrlReg_1;
	setRegisters(CTRL_REG_1, rtc_data, 1); // und schreiben

	rtc_data[0] = dez_2_bcd(time[0]);
	rtc_data[1] = dez_2_bcd(time[1]);
	rtc_data[2] = time[2];						// Foramt DOW muss dezimal bleiben !

	setRegisters(ALARM_REG_B, rtc_data, 3); // und schreiben
}

// ----------------------------------------------------------------------------
// Holt nach einem Interrupt die Flags aus CTRL_REG_2 ab,
// lscht die Flags in Register CTRL_REG_2,
// und gibt die drei Flags freigestellt zurck.
uint8_t rtc_getAlarmFlags(void)
{
	uint8_t i;

	rtc_data[0] = RTC_TWI_ADR;				// TWI-Adresse
	TWI_MA_Read(rtc_data, 2);				// 1 Byte (+ TWI-Adresse) lesen

	if (rtc_data[1] & (1<<XSTP)) error(QUIT);

	i = rtc_data[1] & 0x07;					// Flags maskieren und zwischenspeichern
	rtc_data[0] = (1<<HOUR24 | 1<<CLEN);// nur Flags HOUR24 und CLEN gesetzt
	setRegisters(CTRL_REG_2, rtc_data, 1); // alle Alarmflags lschen

	return i;
}

// ----------------------------------------------------------------------------
// Lscht das EnableFlag fr den Alarm_A im CTRL_REG_1,
// damit wird der Alarm_A deaktiviert.
void rtc_clrAlarm_A(void)
{
	ctrlReg_1 &= (~(1<<AALE));			// der Wert von CTRL_REG_1 steht im RAM
	rtc_data[0] = ctrlReg_1;

	setRegisters(CTRL_REG_1, rtc_data, 1);
}

// ----------------------------------------------------------------------------
// Lscht das EnableFlag fr den Alarm_B im CTRL_REG_1,
// damit wird der Alarm_B deaktiviert.
void rtc_clrAlarm_B(void)
{
	ctrlReg_1 &= (~(1<<BALE));			// der Wert von CTRL_REG_1 steht im RAM
	rtc_data[0] = ctrlReg_1;
	setRegisters(CTRL_REG_1, rtc_data, 1);
}

// -----------------------------------------------------------------------------
// Setzt bzw. lscht den Periodic Interrupt im CTRL_REG_1
// 0   = lschen
// 1-7 = einstellen
void rtc_setPeriodicIRQ(uint8_t i)
{
	i &= 0x07;	// auf die zustndigen Bits reduzieren

	ctrlReg_1 &= (~(0x07));	// ggf. gesetzte Bits lschen
	ctrlReg_1 |= i;			// und neue Einstellung einkopieren
	rtc_data[0] = ctrlReg_1;

	setRegisters(CTRL_REG_1, rtc_data, 1);
}

// -----------------------------------------------------------------------------
// Lscht den Periodic Interrupt im CTRL_REG_1
// Die Funktion kann auch eingespart werden,
// denn rtc_setPeriodicIRQ mit dem Parameter 0 bewirkt dasselbe !
void rtc_clrPeriodicIRQ(void)
{
	ctrlReg_1 &= (~(0x07));
	rtc_data[0] = ctrlReg_1;

	setRegisters(CTRL_REG_1, rtc_data, 1);
}

// ----------------------------------------------------------------------------
// zweizeilig Ausgabe (je 80 Zeichen) der Registerinhalte
// 1.Zeile -> Registernummern hexadezimal
// 2.Zeile -> Werte hexadezimal
void rtc_printRegisters(void)
{
	uint8_t i, j;

	twibuf[0] = RTC_TWI_ADR;
	TWI_MA_Read(twibuf, 17);  // 16 Byte (+ TWIAdresse) lesen

	// 1. Zeile - Registernummer ausgeben
	j = 0x0F;
	put_c(13);
	for (i = 1; i < 17; i++)
	{
		print_hexbyte(j++);
		j &= 0x0F;
	}

	// 2. Zeile - Registerinhalte ausgeben
	put_c(13);
	for (i = 1; i < 17; i++)
	{
		print_hexbyte(twibuf[i]);
	}
	put_c(13);
	put_c(13);
}

// ----------------------------------------------------------------------------
// lokal -> Wandelt eine BCD-Zahl in eine Dezimal-Zahl um
uint8_t bcd_2_dez(uint8_t bcd)
{
	uint8_t dez;
	dez = (bcd >> 4) * 10;
	dez += (bcd & 0x0F);
	return dez;
}

// ----------------------------------------------------------------------------
// lokal -> Wandelt eine Dezimal-Zahl in eine BCD-Zahl um
uint8_t dez_2_bcd(uint8_t dez)
{
	uint8_t bcd;

	bcd = (dez / 10) << 4;
	bcd += dez % 10;
	return bcd;
}

// ----------------------------------------------------------------------------
// Beschreibt die Register beginnend mit 'startReg' mit 'cnt' Bytes aus Array 'values'.
// Schreibt nach der TWI-Adresse die Registeradresse (ins upper-nibble verschoben) und
// dann die nacheinander die Registerwerte.
void setRegisters( uint8_t startReg, uint8_t * values, uint8_t cnt)
{
	uint8_t i, j;

	twibuf[0] = RTC_TWI_ADR;
	twibuf[1] = (startReg << 4);		// Registernummer ins upper-nibble

	i = 0;
	j = 2;

	while (i < cnt) twibuf[j++] = values[i++];// die Daten anhngen

	TWI_MA_Write(twibuf, cnt + 2);
}

// ----------------------------------------------------------------------------
// local -> Holt nur die beiden ControlRegister ab.
// Beginnt mit dem Lesen bei Adresse 0x0E (nicht bei 0x0F) !
// Dazu ist nach dem Schreiben der gewnschten RegisterAdresse
// ein Repeated Start erforderlich !
// Die Registerwerte stehen dann in twibuf[1] und twibuf[2]
void getCtrlRegisters(uint8_t *data)
{
	twibuf[0] = RTC_TWI_ADR;
	twibuf[1] = (CTRL_REG_1 << 4);

	TWI_MA_RepeatedStart();			// nach dem nchsten TWI_Write kein TWI-Stop ausfhren !
	TWI_MA_Write(twibuf, 2);
	TWI_MA_Read(twibuf, 3);

	data[0] = twibuf[1];
	data[1] = twibuf[2];

/*
	put_c(13);
	puts_p(PSTR("ControlRegister: "));
	print_hexbyte(twibuf[1]);
	put_c(' ');
	print_hexbyte(twibuf[2]);
	put_c(13);
*/
}

// ----------------------------------------------------------------------------
// local -> Das Flag XSTP war gesetzt, die Zeitinformation der RTC ist fehlerhaft
// Hier muss in geeigneter Form reagiert werden.
void error(uint8_t e)
{
	puts_p(PSTR("Powerdown - "));

	if (e == CONTINUE)				// Beim Programmstart / nach Einschalten der RTC
	{
		clrRegisters();				// alle Register in default-Einstellung bringen
		puts_p(PSTR("RTC reset"));
	}
	else 									// Whrend der Programmausfhrung
	{
		puts_p(PSTR("Error: Exit"));

		_delay_ms(20);
		EIMSK = 0;

		// ------------------------ sleep
		SMCR = (1<<SM1 | 1<<SE);
		asm volatile("sleep");
	}
	put_c(13);
}

// --- [ eof ] ----------------------------------------------------------------
